/*
Module : IconComboBox.cpp
Purpose: Defines the implementation for 3 MFC classes to implement icon selection combo boxes
Created: PJN / 23-08-1998
History: PJN / 12-09-1998 1. Removed some files from the workspace which were not being used.
                          2. Updated workspace to use folders.
                          3. Added AddString, InsertString & DeleteString functions. These 
                          functions ASSERT to ensure that the icon combo box operates correctly
                          4. Added DeleteIcon function. This should be called instead of DeleteString
                          5. Modfied internal implementation of InsertIcon & AddIcon for both classes
         PJN / 27-12-2016 1. Updated copyright details
                          2. Replaced all occurrences of NULL with nullptr throughout the codebase
                          3. Updated the code to clean compile on VC 2010 to VC 2015. The code now
                          requires VC 2010 at a minimum.
                          4. Reworked the CSmallIconComboBox and CLargeIconComboBox classes to use a new 
                          CIconComboBox base class which supports arbitrary sized icons
         PJN / 24-12-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code
                          will now only compile on VC 2017 or later.
         PJN / 02-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 22-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 15-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 14-03-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 14-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 1998 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "IconComboBox.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//////////////////// Locals ///////////////////////////////////////////////////

class CIconComboBoxData
{
public:
//Constructors / Destructors
  CIconComboBoxData() noexcept : m_hIcon{nullptr},
                                 m_pszResourceID{nullptr}
  {
  }
  CIconComboBoxData(const CIconComboBoxData&) = delete;
  CIconComboBoxData(CIconComboBoxData&&) = delete;
  ~CIconComboBoxData()
  {
    if (m_hIcon != nullptr)
    {
      DestroyIcon(m_hIcon);
      m_hIcon = nullptr;
    }
  }

//Methods
  CIconComboBoxData& operator=(const CIconComboBoxData&) = delete;
  CIconComboBoxData& operator=(CIconComboBoxData&&) = delete;

//Member variables
  HICON m_hIcon;
  LPCTSTR m_pszResourceID;
};


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26429)
void DDX_CBIcon(_In_ CDataExchange* pDX, _In_ int nIDC, _Inout_ UINT& nResourceID)
{
#pragma warning(suppress: 26486)
  HWND hWndCtrl{pDX->PrepareCtrl(nIDC)};
#pragma warning(suppress: 26466)
  auto pCombo{static_cast<CIconComboBox*>(CWnd::FromHandlePermanent(hWndCtrl))};
#pragma warning(suppress: 26496)
  AFXASSUME(pCombo != nullptr);
  ASSERT(pCombo->IsKindOf(RUNTIME_CLASS(CIconComboBox)));

  if (pDX->m_bSaveAndValidate)
    nResourceID = pCombo->GetCurSelIconAsInt();
  else
    pCombo->SetCurSelIcon(nResourceID);
}

#pragma warning(suppress: 26429)
void DDX_CBIcon(_In_ CDataExchange* pDX, _In_ int nIDC, _Inout_ LPCTSTR& pszResourceID)
{
#pragma warning(suppress: 26486)
  HWND hWndCtrl{pDX->PrepareCtrl(nIDC)};
#pragma warning(suppress: 26466)
  auto pCombo{static_cast<CIconComboBox*>(CWnd::FromHandlePermanent(hWndCtrl))};
#pragma warning(suppress: 26496)
  AFXASSUME(pCombo != nullptr);
  ASSERT(pCombo->IsKindOf(RUNTIME_CLASS(CIconComboBox)));

  if (pDX->m_bSaveAndValidate)
    pszResourceID = pCombo->GetCurSelIcon();
  else
    pCombo->SetCurSelIcon(pszResourceID);
}


#pragma warning(suppress: 26433 26440 26477)
IMPLEMENT_DYNAMIC(CIconComboBox, CComboBox)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CIconComboBox, CComboBox) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()

CIconComboBox::CIconComboBox() noexcept : m_nCXIcon{-1},
                                          m_nCYIcon{-1}
{
}

void CIconComboBox::PreSubclassWindow()
{
  //Let the parent do its thing
  __super::PreSubclassWindow();

#ifdef _DEBUG
  const DWORD dwStyle{GetStyle()};
  ASSERT(dwStyle & CBS_HASSTRINGS); //combo box must manage the strings
  ASSERT(dwStyle & CBS_OWNERDRAWVARIABLE); //combo box must be owner draw variable
  ASSERT(m_nCYIcon != -1); //Icon size should have been initialized by now
#endif //#ifdef _DEBUG

  //Set the Height of the combo box to just contain one icon plus some some extra space
  SetItemHeight(-1, m_nCYIcon + 2);
}

int CIconComboBox::AddIcon(_In_ LPCTSTR pszResourceID, _In_ HINSTANCE hInstance)
{
  //Validate our parameters
  ASSERT(m_nCXIcon != -1); //Icon size should have been initialized by now
  ASSERT(m_nCYIcon != -1); //Icon size should have been initialized by now

  //Load up the icon
  auto hIcon{static_cast<HICON>(LoadImage(hInstance, pszResourceID, IMAGE_ICON, m_nCXIcon, m_nCYIcon, 0))};
  if (hIcon == nullptr)
    return CB_ERR;

  //Add it to the combo box
  const int nOffset{__super::AddString(_T(""))};
  if ((nOffset == CB_ERR) || (nOffset == CB_ERRSPACE))
    return nOffset;

  //Hive away the info in the item data field
#pragma warning(suppress: 26403 26409)
  auto pData{new CIconComboBoxData};
  pData->m_hIcon = hIcon;
  pData->m_pszResourceID = pszResourceID;
#pragma warning(suppress: 26490)
  if (SetItemDataPtr(nOffset, pData) == CB_ERR)
  {
    __super::DeleteString(nOffset);
    return CB_ERR;
  }

  //Return the offset at which it was added
  return nOffset; //NOLINT(clang-analyzer-cplusplus.NewDeleteLeaks)
}

int CIconComboBox::AddIcon(_In_ UINT nResourceID, _In_ HINSTANCE hInstance)
{
  return AddIcon(MAKEINTRESOURCE(nResourceID), hInstance);
}

int CIconComboBox::InsertIcon(_In_ int nIndex, _In_ LPCTSTR pszResourceID, _In_ HINSTANCE hInstance)
{
  //Validate our parameters
  ASSERT(m_nCXIcon != -1); //Icon size should have been initialized by now
  ASSERT(m_nCYIcon != -1); //Icon size should have been initialized by now

  //Load up the icon
  auto hIcon{static_cast<HICON>(LoadImage(hInstance, pszResourceID, IMAGE_ICON, m_nCXIcon, m_nCYIcon, 0))};
  if (hIcon == nullptr)
    return CB_ERR;

  //Add it to the combo box at the specified offset
  const int nOffset{__super::InsertString(nIndex, _T(""))};
  if ((nOffset == CB_ERR) || (nOffset == CB_ERRSPACE))
    return nOffset;

  //Hive away the info in the item data field
#pragma warning(suppress: 26403 26409)
  auto pData{new CIconComboBoxData};
  pData->m_hIcon = hIcon;
  pData->m_pszResourceID = pszResourceID;
#pragma warning(suppress: 26490)
  if (SetItemDataPtr(nOffset, pData) == CB_ERR)
  {
    __super::DeleteString(nOffset);
    return CB_ERR;
  }

  //Return the offset at which it was added
  return nOffset; //NOLINT(clang-analyzer-cplusplus.NewDeleteLeaks)
}

int CIconComboBox::InsertIcon(_In_ int nIndex, _In_ UINT nResourceID, _In_ HINSTANCE hInstance)
{
  return InsertIcon(nIndex, MAKEINTRESOURCE(nResourceID), hInstance);
}

int CIconComboBox::DeleteIcon(_In_ int nIndex)
{
  //just let the parent do its thing
  return __super::DeleteString(nIndex);
}

LPCTSTR CIconComboBox::GetCurSelIcon() const
{
  const int nSel{GetCurSel()};
  if (nSel == CB_ERR)
    return nullptr;

#pragma warning(suppress: 26490)
  const CIconComboBoxData* pData{reinterpret_cast<CIconComboBoxData*>(GetItemData(nSel))};
#pragma warning(suppress: 26496)
  AFXASSUME(pData != nullptr);
#pragma warning(suppress: 26487 26489)
  return pData->m_pszResourceID;
}

UINT CIconComboBox::GetCurSelIconAsInt() const
{
  LPCTSTR pszResourceID{GetCurSelIcon()};
  ASSERT(IS_INTRESOURCE(pszResourceID));
#pragma warning(suppress: 26490)
  return reinterpret_cast<ULONG_PTR>(pszResourceID) & 0xFFFF;
}

int CIconComboBox::SetCurSelIcon(_In_ UINT nResourceID)
{
  return SetCurSelIcon(MAKEINTRESOURCE(nResourceID));
}

int CIconComboBox::SetCurSelIcon(_In_ LPCTSTR pszResourceID)
{
  //iterate through all the CB items and set the one which has the corresponding item data
  for (int i=0; i<GetCount(); i++)
  {
#pragma warning(suppress: 26490)
    const CIconComboBoxData* pData{reinterpret_cast<CIconComboBoxData*>(GetItemData(i))};
#pragma warning(suppress: 26496)
    AFXASSUME(pData != nullptr);
#pragma warning(suppress: 26489)
    if (pData->m_pszResourceID == pszResourceID)
      return SetCurSel(i);
  }
  return CB_ERR;
}

#pragma warning(suppress: 26440)
void CIconComboBox::MeasureItem(LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
  //Validate our parameters
  ASSERT(m_nCYIcon != -1); //Icon size should have been initialized by now

  //Set the height of the each item to be one icon plus some some extra space
  lpMeasureItemStruct->itemHeight = m_nCYIcon + 2;
}

void CIconComboBox::DrawItem(LPDRAWITEMSTRUCT lpDIS)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(lpDIS != nullptr);
  ASSERT(lpDIS->CtlType == ODT_COMBOBOX);
  ASSERT(m_nCXIcon != -1); //Icon size should have been initialized by now
  ASSERT(m_nCYIcon != -1); //Icon size should have been initialized by now

  //Create a CDC from the SDK struct passed in
#pragma warning(suppress: 26486)
  CDC* pDC{CDC::FromHandle(lpDIS->hDC)};
  if (pDC == nullptr)
    return;

  if (lpDIS->itemState & ODS_SELECTED)
  {
    //item is selected
    const COLORREF color{GetSysColor(COLOR_HIGHLIGHT)};
    CBrush br{color};
    pDC->FillRect(&lpDIS->rcItem, &br);
  }
  else
  {
    //Item is not selected
    const COLORREF color{GetSysColor(COLOR_WINDOW)};
    CBrush br{color};
    pDC->FillRect(&lpDIS->rcItem, &br);
  }

  //Draw the icon
#pragma warning(suppress: 26490)
  auto pData{reinterpret_cast<CIconComboBoxData*>(lpDIS->itemData)};
#pragma warning(suppress: 26496)
  AFXASSUME(pData != nullptr);
  ASSERT(lpDIS->itemData != static_cast<DWORD_PTR>(-1));
#pragma warning(suppress: 26486 26489)
  DrawIconEx(pDC->GetSafeHdc(), lpDIS->rcItem.left+1, lpDIS->rcItem.top+1, pData->m_hIcon, m_nCXIcon, m_nCYIcon, 0, nullptr, DI_NORMAL);
}

void CIconComboBox::DeleteItem(LPDELETEITEMSTRUCT lpDeleteItemStruct) 
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(lpDeleteItemStruct != nullptr);

  //Free up any existing HICON's used for drawing
#pragma warning(suppress: 26490)
  auto pData{reinterpret_cast<CIconComboBoxData*>(GetItemData(lpDeleteItemStruct->itemID))};
  if (pData != nullptr)
  {
    //Free up our internal structure which is stored in the item data
#pragma warning(suppress: 26486)
    delete pData;
  }

  //Let the parent do its thing
  __super::DeleteItem(lpDeleteItemStruct);
}

#pragma warning(suppress: 26434 26440)
int CIconComboBox::AddString(_In_ LPCTSTR /*pszString*/)
{
  ASSERT(FALSE); //Use AddIcon not AddString
  return CB_ERR;
}

#pragma warning(suppress: 26434 26440)
int CIconComboBox::InsertString(_In_ int /*nIndex*/, _In_ LPCTSTR /*pszString*/)
{
  ASSERT(FALSE); //Use InsertIcon not InsertString
  return CB_ERR;
}

#pragma warning(suppress: 26434 26440)
int CIconComboBox::DeleteString(_In_ int /*nIndex*/)
{
  ASSERT(FALSE); //Use DeleteIcon not DeleteString
  return CB_ERR;
}

void CIconComboBox::SetIconSize(_In_ int nCX, _In_ int nCY) noexcept
{
  m_nCXIcon = nCX;
  m_nCYIcon = nCY;
}

void CIconComboBox::SetUseSmallIconSize() noexcept
{
  SetIconSize(GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON));
}

void CIconComboBox::SetUseLargeIconSize() noexcept
{
  SetIconSize(GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON));
}


#pragma warning(suppress: 26433 26440 26477)
IMPLEMENT_DYNAMIC(CSmallIconComboBox, CIconComboBox)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CSmallIconComboBox, CIconComboBox) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()

CSmallIconComboBox::CSmallIconComboBox() noexcept
{
  SetUseSmallIconSize();
}


#pragma warning(suppress: 26433 26440 26477)
IMPLEMENT_DYNAMIC(CLargeIconComboBox, CIconComboBox)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CLargeIconComboBox, CIconComboBox) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()

CLargeIconComboBox::CLargeIconComboBox() noexcept
{
  SetUseLargeIconSize();
}
